import java.awt.Graphics;
import java.awt.event.MouseEvent;
import java.awt.event.MouseListener;
import java.io.FileInputStream;
import java.io.FileNotFoundException;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.ObjectInputStream;
import java.io.ObjectOutputStream;

import javax.swing.JFileChooser;
import javax.swing.JPanel;

/*
 The canvas section of the Paint program
 this panel is the area where all the shapes are drawn
 shapes are placed into a list to be drawn when paint is called
 also has an undo/redo stack for indecisive artists
 This class also has a reference back to the main program to get the current brush
 */
public class DrawPanel extends JPanel implements MouseListener {
	private ListArray<Shape> layers = new ListArray<Shape>();// shapes to be drawn
	private StackArray<Shape> undo = new StackArray<Shape>();// action history
	private StackArray<Shape> redo = new StackArray<Shape>();// undo history
	private JavaPaint owner;// reference to the main container
	private Shape current;// holds the current brush so I dont have to call getBrush 200 times
	private Point start;// holds starting XY coordinates for mouse press and release
	private Point end;// holds ending XY coordinates for mouse press and release

	// main constructor
	// takes a reference to the main program
	public DrawPanel(JavaPaint ref) {
		super();
		this.setVisible(true);
		owner = ref;// refrence back to JavaPaint
		addMouseListener(this);// mouselistener to get XY coordinates
	}

	// displaying drawn shapes
	public void paint(Graphics g) {
		super.paint(g);
		for (int i = 0; i < layers.size(); i++) {
			layers.get(i).draw(g);// draws drawn shapes in order
		}
		repaint();// updates the panel as more shapes are drawn
	}

	//Saving pictures to disk
	//saves the list of shapes to the disk
	public void save() {
		JFileChooser savePrompt = new JFileChooser();
		savePrompt.showDialog(null, "Save File");//user specifies save location

		try {//sets up the output stream
			ObjectOutputStream output = new ObjectOutputStream(new FileOutputStream(savePrompt.getSelectedFile()));
			output.writeObject(layers);//writes the list to a file
			output.close();//closes the stream
		} catch (Exception oe) {
			throw new JavaPaintException("could not save");
		}
	}

	//loads a picture from disk
	//gets the list of shapes and layers refer to it
	public void load() {
		JFileChooser prompt = new JFileChooser();
		prompt.showDialog(null, "Open File");//user specifies which file to load

		try {//sets up the inputstream
			ObjectInputStream input = new ObjectInputStream(new FileInputStream(prompt.getSelectedFile()));
			Object temp = input.readObject();//gets the object
			if(temp instanceof ListArray) layers=(ListArray<Shape>)temp;//checks compatibility and casts
			else throw new JavaPaintException("file is not a JavaPaint picture");
			input.close();//closes the stream
		} catch (Exception ie) {
			throw new JavaPaintException("could not load");
		}
	}

	// sorts the shapes that has been drawn
	// by area in decending order
	public void sort() {
		layers.sort();
	}

	// undoes the last shape drawn
	// grabs the last shape drawn from the undo stack
	// searches the layers list for that shape
	// removes from the list and puts onto the redo stack
	public void undo() {
		if (undo.isEmpty())
			throw new JavaPaintException("no more actions to undo");
		redo.push(layers.remove(undo.pop()));
	}

	// undoes last shape undone
	// grabs the last shape undone and held in a temp variable
	// places the temp variable into the layer list and the stack
	public void redo() {
		if (undo.isEmpty())
			throw new JavaPaintException("no more actions to redo");
		Shape temp = redo.pop();// holds the shape popped
		layers.add(temp);// adds to the drawing list
		undo.push(temp);// adds to the undo stack
	}

	// Calculates the distance between 2 points
	// used to determine the size of the shape to be drawn
	private int distance(int a, int b) {
		return Math.abs(b - a);
	}

	@Override
	// stamps shapes onto the canvas w/ a default size of 50
	// and at the current mouse location
	public void mouseClicked(MouseEvent e) {
		System.out.println("DrawPanel clicked at " + e.getX() + "," + e.getY());// debug
		if (owner.getBrush() != null && !(owner.getBrush() instanceof Line)) {// null/line check, line should only draw on mouseRelease
			current = owner.getBrush();// gets the brush
			current.setX(e.getX());// sets the XY
			current.setY(e.getY());
			current.setSize(50);// default size
		}
	}

	@Override
	// the fun mouse event
	// dynamically sizes shapes being drawn
	// from the starting location of the mouse press to the mouse release
	// which finishes the drawing
	public void mousePressed(MouseEvent e) {
		System.out.println("DrawPanel pressed at " + e.getX() + "," + e.getY());// debug
		if (owner.getBrush() != null) {// null check
			start = new Point(e.getX(), e.getY());// stores the starting XY to be compared in mouseReleased()
			current = owner.getBrush();// gets the brush

			current.setX(start.x);// set XY
			current.setY(start.y);
		}
		System.out.println(layers.toString());// debug
	}

	@Override
	// this actually draws the shapes from mousePressed()
	public void mouseReleased(MouseEvent e) {
		System.out.println("DrawPanel released at " + e.getX() + "," + e.getY());// debug

		if (owner.getBrush() != null) {// null check
			end = new Point(e.getX(), e.getY());// gets XY
			int deltaX = distance(start.x, end.x);// calculates the size of the
													// shape to be drawn
			int deltaY = distance(start.y, end.y);

			if (current instanceof Line) {// line check, sizing is handled differently compared to square and circle
				Line that = (Line) current;
				that.setX2(end.x);// sets ending coordinate
				that.setY2(end.y);

			} else {
				if (end.x < start.x)// checks if the mouse moved left
					current.setX(end.x);// sets the endpoint as the start point
				if (end.y < start.y)// checks if the mouse moved up
					current.setY(end.y);// sets the endpoint as the start point

				// checks instanceof to use the right setter
				if (current instanceof Square) {
					Square that = (Square) current;// casts to a square
					that.setW(deltaX);// sets width and height
					that.setH(deltaY);
				} else {
					Circle that = (Circle) current;// casts to a circle
					that.setW(deltaX);// sets width and height
					that.setH(deltaY);
				}
			}

			layers.add(current);// draws the shape
			undo.push(current);// adds to the undo stack
		}
		System.out.println(layers.toString());// debug
	}

	@Override
	public void mouseEntered(MouseEvent e) {
	}

	@Override
	public void mouseExited(MouseEvent e) {
	}
}